/*
 *  $Id: adaptation.c,v 1.49 2001/01/04 18:13:15 ajung Exp $
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 2000 by Siemens AG, Munich, Germany.
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at http://www.sctp.de which should
 * be used for any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          Michael.Tuexen@icn.siemens.de
 *          ajung@exp-math.uni-essen.de
 *
 * Purpose: The adaption-module encapsulates the socket-interface.
 *          The adaption-module also handles timers in the SCTP and its ULP.
 */

#include "globals.h"
/* for struct timer_list */
#include "timer_list.h"
/* for a simple sctp callback function */
#include "sctp.h"

#include "adaptation.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <netinet/in_systm.h>
#include <netinet/ip.h>    
#include <netinet/ip_icmp.h>
#include <netdb.h>
#include <arpa/inet.h>          /* for inet_ntoa() under both SOLARIS/LINUX */

#include <sys/errno.h>
#include <errno.h>
#ifdef HAVE_SYS_POLL_H
#include <sys/poll.h>
#else
#define POLLIN     0x001
#define POLLPRI    0x002
#define POLLOUT    0x004
#define POLLERR    0x008

struct pollfd {
	int fd;	
    short int events;
	short int revents;
};
#endif 

#ifndef HAVE_POLL
/* a simple poll implementation based on select */
int poll(struct pollfd *fdlist, long unsigned int count, int time)
{
  struct timeval timeout, *to;
  fd_set readfdset, writefdset;
  long unsigned int i;
  int ret, fdcount = 0;
  int tsize = getdtablesize();


  if (time < 0)
    to = NULL;
  else {
    to = &timeout;
    timeout.tv_sec = time / 1000;
    timeout.tv_usec = (time % 1000) * 1000;
  }

  FD_ZERO(&readfdset);
  FD_ZERO(&writefdset);
  for (i = 0; i < count; i++) {
    if (fdlist[i].fd < 0) continue;
    if (fdlist[i].events & POLLIN)  FD_SET(fdlist[i].fd, &readfdset);
    if (fdlist[i].events & POLLOUT) FD_SET(fdlist[i].fd, &writefdset);
    fdcount++;
  }

  if (fdcount == 0)
    return(0);

  for (i = 0; i < count; i++) fdlist[i].revents = 0;

  if ((ret = select(tsize, &readfdset, &writefdset, NULL, to)) == -1)
    return(-1);

  for (i = 0; i < count; i++) {
    if (FD_ISSET(fdlist[i].fd, &readfdset) && (fdlist[i].events & POLLIN))
        fdlist[i].revents |= POLLIN;
    if (FD_ISSET(fdlist[i].fd, &writefdset) && (fdlist[i].events & POLLOUT))
        fdlist[i].revents |= POLLOUT;
  }
  return(ret);
}
#endif

/* THIS IS HIGHLY SYSTEM SPECIFIC, BUT ONLY NEEDED FOR MORE EFFECTIVE
 * ICMP MESSAGE FILTERING
 * WE COULD ALSO FILTER THIS IN OUR USER LEVEL PROGRAM.....
 * WE WOULD HAVE TO IGNORE ALL MESSAGES BUT "Destination Unreachable"
 */
#ifndef ICMP_FILTER
#define ICMP_FILTER                     1
#endif

#ifdef HAVE_IPV6
#include <netinet/icmp6.h>
#endif

#ifndef LINUX
 /* ICMP messages */
#define ICMP_ECHOREPLY          0 /* Echo Reply                   */
#define ICMP_DEST_UNREACH       3 /* Destination Unreachable      */
#define ICMP_SOURCE_QUENCH      4 /* Source Quench                */
#define ICMP_REDIRECT           5 /* Redirect (change route)      */
#define ICMP_ECHO               8 /* Echo Request                 */
#define ICMP_TIME_EXCEEDED      11 /* Time Exceeded                */
#define ICMP_PARAMETERPROB      12 /* Parameter Problem            */
#define ICMP_TIMESTAMP          13 /* Timestamp Request            */
#define ICMP_TIMESTAMPREPLY     14 /* Timestamp Reply              */
#define ICMP_INFO_REQUEST       15 /* Information Request          */
#define ICMP_INFO_REPLY         16 /* Information Reply            */
#define ICMP_ADDRESS            17 /* Address Mask Request         */
#define ICMP_ADDRESSREPLY       18 /* Address Mask Reply           */
#define NR_ICMP_TYPES           18

 /* Codes for UNREACH. */
#define ICMP_NET_UNREACH        0 /* Network Unreachable          */
#define ICMP_HOST_UNREACH       1 /* Host Unreachable             */
#define ICMP_PROT_UNREACH       2 /* Protocol Unreachable         */
#define ICMP_PORT_UNREACH       3 /* Port Unreachable             */
#define ICMP_FRAG_NEEDED        4 /* Fragmentation Needed/DF set  */
#define ICMP_SR_FAILED          5 /* Source Route failed          */
#define ICMP_NET_UNKNOWN        6
#define ICMP_HOST_UNKNOWN       7
#define ICMP_HOST_ISOLATED      8
#define ICMP_NET_ANO            9
#define ICMP_HOST_ANO           10
#define ICMP_NET_UNR_TOS        11
#define ICMP_HOST_UNR_TOS       12
#define ICMP_PKT_FILTERED       13 /* Packet filtered */
#define ICMP_PREC_VIOLATION     14 /* Precedence violation */
#define ICMP_PREC_CUTOFF        15 /* Precedence cut off */
#define NR_ICMP_UNREACH         15 /* instead of hardcoding immediate value */

 /* Other defs */
#define IP_MTU_DISCOVER         10
#define IP_PMTUDISC_DONT         0
#define IP_PMTUDISC_WANT         1
#define IP_PMTUDISC_DO           2
#define SOL_RAW                255
#endif

#define POLL_FD_UNUSED 	-1

/* This is actually taken from mdtp_test_app.c ...I won't reinvent the wheel  */
/*   poll_fds. fd will be set to  POLL_FD_UNUSED, if it is not used		*/
/* if X poll_fds are set, they must be in the first X structures 			*/
#define NUM_FDS     10

/* *********************** added code for testing purposes ******************************************/
static int dropflg;             /* 0 for normal behaviour */
static unsigned int number_of_sendevents = 0;
/* ******************************************* up tp here  ******************************************/
static unsigned short udpPort;

static unsigned char rbuf[MAX_MTU_SIZE + 20];


/**
 *  Structure for callback events. The function "action" is called by the event-handler,
 *  when a socket gets data or the ULP is to be notified of some conditions.
 *  The function (action) takes five arguments :
 *  1. socket fd, where event occurred
 *  2. pointer to a (static) buffer where the data sits 	
 *  3. an integer that contains the length of the data
 *  4. pointer to the address where data came from
 *  @param size of the address/sockaddr_in, the origin of the message
 */
struct event_cb
{
    int sfd;
    int event_type;
    /* pointer to possible arguments, associations etc. */
    void *arg1;
    void *arg2;
    void (*action) (int, unsigned char *, int, union sockunion *, int);
};



static struct pollfd poll_fds[NUM_FDS];
static int num_of_fds = 0;

static int sctp_sfd = 0;        /* socket fd for standard SCTP port.... */
static int ulp_sfd = 0;
#ifdef HAVE_IPV6
static int sctpv6_sfd = 0;
static int icmpv6_sfd = 0;
#endif
static int icmp_sfd = 0;        /* socket fd for ICMP messages */

static struct event_cb *event_callbacks[NUM_FDS];


/**
 *  converts address-string (hex for ipv6, dotted decimal for ipv4
 *  to a sockunion structure
 *  @return 0 for success, else -1.
 */
int adl_str2sockunion(guchar * str, union sockunion *su)
{
    int ret;

    memset(su, 0, sizeof(union sockunion));

    ret = inet_aton(str, &su->sin.sin_addr);
    if (ret > 0) {              /* Valid IPv4 address format. */
        su->sin.sin_family = AF_INET;
#ifdef HAVE_SIN_LEN
        su->sin.sin_len = sizeof(struct sockaddr_in);
#endif                          /* HAVE_SIN_LEN */
        return 0;
    }
#ifdef HAVE_IPV6
    ret = inet_pton(AF_INET6, str, &su->sin6.sin6_addr);
    if (ret > 0) {              /* Valid IPv6 address format. */
        su->sin6.sin6_family = AF_INET6;
#ifdef SIN6_LEN
        su->sin6.sin6_len = sizeof(struct sockaddr_in6);
#endif                          /* SIN6_LEN */
        return 0;
    }
#endif                          /* HAVE_IPV6 */
    return -1;
}


const int adl_sockunion2str(union sockunion *su, guchar * buf, size_t len)
{
    if (su->sa.sa_family == AF_INET){
        strncpy(buf, inet_ntoa(su->sin.sin_addr), len);
        return(1);
    }
#ifdef HAVE_IPV6
    else if (su->sa.sa_family == AF_INET6) {
        error_log(ERROR_MAJOR, "TODO : check or change function for IPv6 address conversion !");
        return inet_ntop(AF_INET6, &su->sin6.sin6_addr, buf, len);
    }
#endif                          /* HAVE_IPV6 */
    return 0;
}

boolean adl_equal_address(union sockunion * one, union sockunion * two)
{
    unsigned int count;

    switch (sockunion_family(one)) {
    case AF_INET:
        if (sockunion_family(two) != AF_INET)
            return FALSE;
        return (sock2ip(one) == sock2ip(two));
        break;
#ifdef HAVE_IPV6
    case AF_INET6:
        if (sockunion_family(two) != AF_INET6)
            return FALSE;
        for (count = 0; count < 16; count++)
            if (sock2ip6(one)[count] != sock2ip6(two)[count])
                return FALSE;
        return TRUE;
        break;
#endif
    default:
        error_logi(ERROR_MAJOR, "Address family %d not supported", sockunion_family(one));
        return FALSE;
        break;
    }
}



int adl_open_sctp_socket(int af)
{
    struct protoent *proto;
    int sfd, protoid = 0, ch;

    if (!(proto = getprotobyname("sctp"))) {
        if (Current_error_log_ >= ERROR_WARNING)
            error_log(ERROR_MINOR, "SCTP: unknown protocol sctp - using 132\n");
        protoid = IPPROTO_SCTP;
    }
    if (!protoid)
        protoid = proto->p_proto;
    if ((sfd = socket(af, SOCK_RAW, protoid)) < 0) {
        if (errno == EPERM) {
            if (Current_error_log_ >= ERROR_WARNING)
                error_log(ERROR_FATAL, "SCTP: must run as root or setuid root\n");
        } else
            error_log(ERROR_MAJOR, "SCTP: socket creation failed in adl_open_sctp_socket !");
    }
    ch = 0;
    if (af == AF_INET) {
        if (setsockopt(sfd, IPPROTO_IP, IP_HDRINCL, (char *) &ch, sizeof(ch))
            < 0) {
            error_log(ERROR_FATAL, "setsockopt: IP_HDRINCL");
        }
#if defined (LINUX)
        ch = IP_PMTUDISC_DO;
        if (setsockopt(sfd, IPPROTO_IP, IP_MTU_DISCOVER, (char *) &ch, sizeof(ch)) < 0) {
            error_log(ERROR_FATAL, "setsockopt: IP_PMTU_DISCOVER");
        }
#else
        error_log(ERROR_MINOR, "TODO : PATH MTU Discovery Disabled For Now !!!\n");
#endif
    }
    event_logi(INTERNAL_EVENT_0, "Created raw socket %d with options\n", sfd);
    return (sfd);
}


gint adl_get_sctpv4_socket(unsigned int portnum)
{
    /* this is a static variable ! */
    return sctp_sfd;
}

gint adl_get_udp_socket(unsigned int portnum)
{
    /* this is a static variable ! */
    return ulp_sfd;
}


#ifdef HAVE_IPV6
gint adl_get_sctpv6_socket(unsigned int portnum)
{
    /* this is a static variable ! */
    return sctpv6_sfd;
}
#endif


int adl_open_icmp_socket(void)
{
    struct protoent *proto;
    int ch, sfd, protoid = 0;

    if (!(proto = getprotobyname("icmp"))) {
        error_log(ERROR_FATAL, "SCTP: unknown protocol icmp.\n");
        protoid = 1;
    }

    if (!protoid)
        protoid = proto->p_proto;

    if ((sfd = socket(AF_INET, SOCK_RAW, protoid)) < 0) {
        if (errno == EPERM) {
            error_log(ERROR_FATAL, "SCTP: must run as root or setuid root\n");
        } else
            error_log(ERROR_FATAL, "SCTP: socket creation failed for icmp");
    }

    /* Only get this type of error code. So we do not catch pings for example */
#if defined (LINUX)
    ch = ICMP_DEST_UNREACH;
    if (setsockopt(sfd, SOL_RAW, ICMP_FILTER, (char *) &ch, sizeof(ch)) < 0) {
        error_log(ERROR_FATAL, "setsockopt: ICMP_FILTER failed");
    }
#else
    error_log(ERROR_MINOR, "ICMP Filter Disabled For Now - We Catch All ICMP Messages!!!\n");
#endif
    return (sfd);
}

/**
 * This function creates a UDP socket bound to localhost, for asynchronous
 * interprocess communication with an Upper Layer process.
 * @return the socket file descriptor. Used to register a callback function
 */
int adl_open_ulp_socket(void)
{
    int ch, sfd;
    struct sockaddr_in localhost;

    if ((sfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
        error_log(ERROR_FATAL, "SCTP: socket creation failed for UDP socket for ULP communication");
    }
    localhost.sin_family = AF_INET;
    localhost.sin_port = htons(udpPort);
    /* my localhost address is defined as 127.0.0.1 */
    localhost.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
    ch = bind(sfd, (struct sockaddr *) &localhost, sizeof(struct sockaddr_in));

    event_logiii(VERBOSE,
                 " adl_open_ulp_socket : Create socket %u, binding to port %u, result %d",
                 sfd, udpPort, ch);

    if (ch == 0)
        return (sfd);
    return -1;
}


#ifdef HAVE_IPV6
int adl_open_icmpv6_socket(void)
{
    struct icmp6_filter filter;

    struct protoent *proto;

    int ch, sfd, protoid = 0;

    if (!(proto = getprotobyname("ipv6-icmp"))) {
        error_log(ERROR_MINOR, "SCTP: unknown protocol icmpv6.\n");
        protoid = IPPROTO_ICMPV6;

    }

    if (!protoid)
        protoid = proto->p_proto;

    if ((sfd = socket(AF_INET6, SOCK_RAW, protoid)) < 0) {
        if (errno == EPERM) {
            error_log(ERROR_FATAL, "SCTP: must run as root or setuid root\n");
        } else
            error_log(ERROR_FATAL, "SCTP: socket creation failed for icmpv6");
    }

    /* Only get this type of error code. So we do not catch pings for example */
    ICMP6_FILTER_SETBLOCKALL(&filter);
    ICMP6_FILTER_SETPASS(ICMP6_DST_UNREACH, &filter);
    ICMP6_FILTER_SETPASS(ICMP6_PACKET_TOO_BIG, &filter);
    ICMP6_FILTER_SETPASS(ICMP6_PARAM_PROB, &filter);

    ch = setsockopt(sfd, IPPROTO_ICMPV6, ICMP6_FILTER, &filter, sizeof(struct icmp6_filter));
    if (ch < 0) {
        error_logi(ERROR_MINOR, "ICMP6_FILTER set fail: %s", strerror(errno));
        return ch;
    }
    return (sfd);
}
#endif


/**
 * function to be called when we get a message from a peer sctp instance in the poll loop
 * @param  sfd the socket file descriptor where data can be read...
 * @param  buf pointer to a buffer, where we data is stored
 * @param  len number of bytes to be sent, including the ip header !
 * @param  address, where data goes from
 * @param	dest_len size of the address
 * @return returns number of bytes actually sent, or error
 */
int adl_send_message(int sfd, void *buf, int len, union sockunion *dest)
{
    int txmt_len = 0, dest_len;
    guchar hostname[MAX_MTU_SIZE];

    event_logi(VVERBOSE, "adl_send_message : drop flag==%d", dropflg);

    switch (sockunion_family(dest)) {

    case AF_INET:
  /* *********************** added code for testing purposes ******************************************/
        if (dropflg != 0) {
            number_of_sendevents++;
            if ((number_of_sendevents % dropflg) == 0) {
                event_logi(INTERNAL_EVENT_0,
                           "AF_INET : adl_send_message : dropping data (%d bytes)", len);
                return len;
            }
        }
  /* ******************************************* up tp here  ******************************************/
        event_logiii(VERBOSE,
                     "AF_INET : adl_send_message : sfd : %d, len %d, destination : %s",
                     sfd, len, inet_ntoa(dest->sin.sin_addr));
        dest_len = sizeof(struct sockaddr_in);
        txmt_len = sendto(sfd, buf, len, 0, (struct sockaddr *) &(dest->sin), dest_len);
        if (txmt_len < 0) {
            perror("AF_INET : Sendto returned error : ");
            error_logi(ERROR_MAJOR, "AF_INET : sendto()=%d, retrying !", txmt_len);
            txmt_len = sendto(sfd, buf, len, 0, (struct sockaddr *) &(dest->sin), dest_len);
            error_logi(ERROR_MAJOR, "AF_INET : Second sendto() returned %d !", txmt_len);
        }
        break;
#ifdef HAVE_IPV6
    case AF_INET6:
  /* *********************** added code for testing purposes ******************************************/
        if (dropflg != 0) {
            number_of_sendevents++;
            if ((number_of_sendevents % dropflg) == 0) {
                event_logi(INTERNAL_EVENT_0,
                           "AF_INET6 : adl_send_message : dropping data (%d bytes)", len);
                return len;
            }
        }
  /* ******************************************* up tp here  ******************************************/
        inet_ntop(AF_INET6, sock2ip6(dest), hostname, MAX_MTU_SIZE);

        event_logiii(VERBOSE,
                     "adl_send_message : sfd : %d, len %d, destination : %s", sfd, len, hostname);

        dest_len = sizeof(struct sockaddr_in6);
        txmt_len = sendto(sfd, buf, len, 0, &(dest->sin6), dest_len);
        if (txmt_len < 0) {
            perror("AF_INET6 : Sendto returned error : ");
            error_logi(ERROR_MAJOR, "AF_INET6 : sendto()=%d, retrying !", txmt_len);
            txmt_len = sendto(sfd, buf, len, 0, &(dest->sin6), dest_len);
            error_logi(ERROR_MAJOR, "AF_INET6 : Second sendto() returned %d !", txmt_len);
        }

        break;

#endif

    default:
        error_logi(ERROR_MAJOR,
                   "adl_send_message : Adress Family %d not supported here",
                   sockunion_family(dest));
        txmt_len = -1;
    }
    return txmt_len;
}

/**
 * function to assign an event mask to a certain poll
 */
void assign_poll_fd(int fd_index, int sfd, int event_mask)
{
    if (fd_index > NUM_FDS)
        error_log(ERROR_FATAL, "FD_Index bigger than NUM_FDS ! bye !\n");

    poll_fds[fd_index].fd = sfd; /* file descriptor */
    poll_fds[fd_index].events = event_mask;
}


/**
 * remove a sfd from the poll_list, and shift that list to the left
 * @return number of sfd's removed...
 */
int remove_poll_fd(int sfd)
{
    int i, tmp, counter = 0;
    for (i = 0, tmp = 0; i < NUM_FDS; i++, tmp++) {
        if (tmp < NUM_FDS) {
            poll_fds[i].fd = poll_fds[tmp].fd;
            poll_fds[i].events = poll_fds[tmp].events;
            event_callbacks[i] = event_callbacks[tmp];
        } else {
            poll_fds[i].fd = POLL_FD_UNUSED;
            poll_fds[i].events = 0;
            free(event_callbacks[i]);
            event_callbacks[i] = NULL;
        }
        if (poll_fds[i].fd == sfd) {
            tmp = i + 1;
            if (tmp < NUM_FDS) {
                poll_fds[i].fd = poll_fds[tmp].fd;
                poll_fds[i].events = poll_fds[tmp].events;
                event_callbacks[i] = event_callbacks[tmp];
            } else {
                poll_fds[i].fd = POLL_FD_UNUSED;
                poll_fds[i].events = 0;
                free(event_callbacks[i]);
                event_callbacks[i] = NULL;
            }
            counter++;
            num_of_fds -= 1;
        }
    }
    return (counter);
}

void
adl_socket_error(gint socket_fd,
                 unsigned char *buffer,
                 int bufferLength, union sockunion *fromAddress, int fromAddressLength)
{
    error_logi(ERROR_MAJOR, "ADL: Socket error occurred when reading from FD %d",socket_fd);
}


/**
 * function to register a file descriptor, that gets activated for certain read/write events
 * when these occur, the specified callback funtion is activated and passed the parameters
 * that are pointed to by the event_callback struct
 */
int
adl_register_fd_cb(int sfd, int event_mask,
                   void (*action) (int, unsigned char *, int, union sockunion *, int))
{


    if (num_of_fds < NUM_FDS) {
        assign_poll_fd(num_of_fds, sfd, event_mask);
        event_callbacks[num_of_fds] = malloc(sizeof(struct event_cb));
        if (!event_callbacks[num_of_fds])
            error_log(ERROR_FATAL, "Could not allocate memory in  register_fd_cb \n");
        event_callbacks[num_of_fds]->sfd = sfd;
        event_callbacks[num_of_fds]->event_type = 0;
        event_callbacks[num_of_fds]->action = action;
        num_of_fds++;
        return num_of_fds;
    } else
        return (-1);
}


/**
 * function to be called when we get a message from a peer sctp instance in the poll loop
 * @param  sfd the socket file descriptor where data can be read...
 * @param  dest pointer to a buffer, where we can store the received data
 * @param  maxlen maximum number of bytes that can be received with call
 * @param  address, where we got the data from
 * @param	from_len size of the address
 * @return returns number of bytes received with this call
 */
int get_message(int sfd, void *dest, int maxlen, union sockunion *from, int *from_len)
{
    int len;

    len = recvfrom(sfd, dest, maxlen, 0, (struct sockaddr *) from, from_len);
    if (len < 0)
        error_log(ERROR_FATAL, "recvfrom  failed in get_message(), aborting !");

    return len;
}

/**
 * this function is responsible for calling the callback functions belonging
 * to all of the file descriptors that have indicated an event !
 * TODO : check handling of POLLERR situation
 * TODO : check handling of several read events at once
 * @param num_of_events  number of events indicated by poll()
 */
void dispatch_event(int num_of_events)
{
    int i = 0, j = 0;
    int length, src_len;
    union sockunion src;
    struct sockaddr_in *src_in;
#ifdef HAVE_IPV6
    guchar hostname[MAX_MTU_SIZE];
#endif

#if !defined (LINUX)
    struct ip *iph;
#else
    struct iphdr *iph;
#endif
    int hlen;

    if (num_of_events > 1)
        error_log(ERROR_MINOR, "Several events in one poll !");

    for (j = 0; j < num_of_events; j++) {
        for (i = 0; i < num_of_fds; i++) {
            if (poll_fds[i].revents) {

                if (poll_fds[i].revents & POLLERR) {
                    /* We must have specified this callback funtion for treating/logging the error */
                    error_log(ERROR_MINOR, "Poll Error Condition ");
                    hlen = 0;
                    length = 0;
                    src_len = 0;

                    (*(event_callbacks[i]->action)) (poll_fds[i].fd, &rbuf[hlen], length, NULL, src_len);
                    break;
                }

                /* FOR THE BAKEOFF -> catch events from stdin */
                if (poll_fds[i].fd == 0) {
                    hlen = 0;
                    length = 0;
                    src_len = 0;

                    (*(event_callbacks[i]->action)) (poll_fds[i].fd,
                                                     &rbuf[hlen], length, NULL, src_len);
                    break;
                }

                src_len = sizeof(src);

                if ((poll_fds[i].revents & POLLPRI)
                    || (poll_fds[i].revents & POLLIN)) {

                    length = get_message(poll_fds[i].fd, rbuf, MAX_MTU_SIZE, &src, &src_len);

                    if (poll_fds[i].fd == ulp_sfd && ulp_sfd != 0) {
                        event_logi(VERBOSE, "Message %d bytes - Activating UDP callback", length);
                        (*(event_callbacks[i]->action)) (poll_fds[i].fd,
                                                         rbuf, length, &src, src_len);
                    } else {

                        switch (sockunion_family(&src)) {

                        case AF_INET:
                            src_in = (struct sockaddr_in *) &src;
                            src_len = sizeof(struct sockaddr_in);
                            event_logi(VERBOSE,
                                       "Message from %s -> activating callback",
                                       inet_ntoa(src_in->sin_addr));
#if !defined (LINUX)
                            iph = (struct ip *) rbuf;
#else
                            iph = (struct iphdr *) rbuf;
#endif
#if !defined (LINUX)
                            hlen = iph->ip_hl << 2;
#else
                            hlen = iph->ihl << 2;
#endif
                            if (length < hlen) {
                                error_logii(ERROR_MINOR,
                                            "dispatch_event : packet too short (%d bytes) from %s",
                                            length, inet_ntoa(src_in->sin_addr));
                                return;
                            }
                            length -= hlen;

                            (*(event_callbacks[i]->action))
                                (poll_fds[i].fd, &rbuf[hlen], length, &src, src_len);

                            break;
#ifdef HAVE_IPV6
                        case AF_INET6:

                            inet_ntop(AF_INET6, sock2ip6(&src), hostname, MAX_MTU_SIZE);

                            event_logii(VERBOSE,
                                        "Message from %s (%d bytes) -> activating callback",
                                        hostname, length);


                            (*(event_callbacks[i]->action))
                                (poll_fds[i].fd, rbuf, length, &src, src_len);

                            break;

#endif                          /* HAVE_IPV6 */
                        default:
                            error_logi(ERROR_MAJOR,
                                       "Unsupported Address Family Type ", sockunion_family(&src));
                            break;

                        }
                    }           /* else */
                }

            }                   /*    if (poll_fds[i].revents) */
            poll_fds[i].revents = 0;

        }                       /*   for(i = 0; i < num_of_fds; i++) */
    }                           /*   for(j = 0; i < num_of_events; j++) */
}


/**
 * function calls the respective callback funtion, that is to be executed as a timer
 * event, passing it two arguments
 */
void dispatch_timer()
{
    int tid, result;
    AlarmTimer event;
    if (timer_list_empty())
        return;

    result = get_msecs_to_nexttimer();
    if (result == 0) {
        result = get_next_event(&event);
        if (!result)
            error_log(ERROR_FATAL, "get_next_event did not return 0\n");
        tid = event.timer_id;
        (*(event.action)) (tid, event.arg1, event.arg2);
        result = remove_item(tid, &event);
        if (result)
            error_log(ERROR_FATAL, "remove_item did not return 0\n");
    }
    return;
}


/**
 *  handler for ICMP events.....
 *  TODO :  check to which association in the list the NeedFragmentation Message
 *          belongs, and adjust the respective MTU setting accordingly
 *  TODO :  activate/deactivate the PMTU-Discovery every once in a while....
 */
void adl_icmp_cb(int sfd, unsigned char *buffer, int buflen, union sockunion *src, int src_len)
{
    /* buflen bytes read ! */
#if !defined (LINUX)
    struct icmp *icp;
    icp = (struct icmp *) buffer;
#else
    struct icmphdr *icp;
    icp = (struct icmphdr *) buffer;
#endif

#if !defined (LINUX)
    if (icp->icmp_type == ICMP_ECHOREPLY) {
#else
    if (icp->type == ICMP_ECHOREPLY) {
#endif
        event_log(EXTERNAL_EVENT, "Got icmp packet with ECHO REPLY");
#if !defined (LINUX)
    } else if (icp->icmp_type == ICMP_DEST_UNREACH) {
#else
    } else if (icp->type == ICMP_DEST_UNREACH) {
#endif
        event_log(EXTERNAL_EVENT, "ICMP ERROR TYPE : Destination Unreachable");
#if !defined (LINUX)
        switch (icp->icmp_code) {
#else
        switch (icp->code) {
#endif
        case ICMP_NET_UNREACH:
            event_log(EXTERNAL_EVENT, "Destination Net Unreachable");
            break;
        case ICMP_HOST_UNREACH:
            event_log(EXTERNAL_EVENT, "Destination Host Unreachable");
            break;
        case ICMP_PROT_UNREACH:
            event_log(EXTERNAL_EVENT, "Destination Protocol Unreachable");
            break;
        case ICMP_PORT_UNREACH:
            event_log(EXTERNAL_EVENT, "Destination Port Unreachable");
            break;
        case ICMP_FRAG_NEEDED:
            event_log(EXTERNAL_EVENT, "frag needed and DF set");
#if defined (SOLARIS)
            /* Here we use ih_idseq struct, but it should contain the same
             * data as the fragment report..... */
            event_logi(EXTERNAL_EVENT, "MTU : %d ", ntohs(icp->icmp_hun.ih_idseq.icd_seq));
#elif defined (LINUX)
            event_logi(EXTERNAL_EVENT, "MTU : %d ", ntohs(icp->un.frag.mtu));
#endif
            /* TODO : HERE WE MAY HAVE TO ADJUST THE NEW PATH MTU */
            break;
        case ICMP_SR_FAILED:
            event_log(EXTERNAL_EVENT, "Source Route Failed");
            break;
        case ICMP_NET_UNKNOWN:
            event_log(EXTERNAL_EVENT, "Network Unknown");
            break;
        case ICMP_HOST_UNKNOWN:
            event_log(EXTERNAL_EVENT, "Host Unknown");
            break;
        case ICMP_HOST_ISOLATED:
            event_log(EXTERNAL_EVENT, "Host Isolated");
            break;
        case ICMP_NET_UNR_TOS:
            event_log(EXTERNAL_EVENT, "Destination Network Unreachable At This TOS");
            break;
        case ICMP_HOST_UNR_TOS:
            event_log(EXTERNAL_EVENT, "Destination Host Unreachable At This TOS");
            break;
        case ICMP_PKT_FILTERED:
            event_log(EXTERNAL_EVENT, "Packet Filtered");
            break;
        case ICMP_PREC_VIOLATION:
            event_log(EXTERNAL_EVENT, "Precedence Violation");
            break;
        default:
            event_log(EXTERNAL_EVENT, "Unknown ICMP ERROR CODE");
            break;
        }
    } else {
        event_log(EXTERNAL_EVENT, "unknown ICMP TYPE");
    }
    return;
}


#ifdef HAVE_IPV6
/**
 *  handler for ICMPv6 events.....THIS NEEDS TO BE FIXED !!!!!!!!!!!!!!!!!!
    THIS FUNCTION IS JUST A DUMMY FOR NOW !!! UNTIL WE HAVE A 6BONE NET UP and RUNNING
 *  TODO :  check to which association in the list the NeedFragmentation Message
 *          belongs, and adjust the respective MTU setting accordingly
 */
void adl_icmpv6_cb(int sfd, unsigned char *buffer, int buflen, union sockunion *src, int src_len)
{
    struct icmp6_hdr *icp;
    int i;
    guchar hostname[MAX_MTU_SIZE];
    guchar *cp, *dp;
    struct ip6hdr *ip;
    int hlen, dupflag;
    /* there might be an ip header first ?! */
    icp = (struct icmp6_hdr *) buffer;

    inet_ntop(AF_INET6, sock2ip6(src), hostname, MAX_MTU_SIZE);

    event_logii(EXTERNAL_EVENT, "icmpv6 packet (%d bytes) from %s", buflen, hostname);

    /* Now the ICMP part - THIS MAY STILL BE BUGGY !!!!!!!!!!!!!!!!!!!!!!!! */

    if (ntohs(icp->icmp6_type) == ICMP6_ECHO_REPLY) {
        event_log(VERBOSE, "Got icmpv6 packet with ECHO REPLY ");
        return;
    } else if (ntohs(icp->icmp6_type) == ICMP6_DST_UNREACH) {
        switch (ntohs(icp->icmp6_code)) {
        case ICMP6_DST_UNREACH_NOROUTE:
            event_log(VERBOSE, "No route\n");
            break;
        case ICMP6_DST_UNREACH_NOTNEIGHBOR:
            event_log(VERBOSE, "Not neighbour\n");
            break;
        case ICMP6_DST_UNREACH_ADDR:
            event_log(VERBOSE, "Address Unreachable\n");
            break;
        case ICMP6_DST_UNREACH_NOPORT:
            event_log(VERBOSE, "Destination Port Unreachable\n");
            break;
        default:
            event_logi(VERBOSE, "Dest Unreachable, Unknown Code: %d\n", icp->icmp6_code);
            break;
        }
    } else {
        event_logi(EXTERNAL_EVENT, "Got icmp packet with type %d\n", icp->icmp6_type);
        event_logi(EXTERNAL_EVENT, "testraw : Unknown Code: %d\n", icp->icmp6_code);
    }
}
#endif


void adl_add_msecs_totime(struct timeval *t, unsigned int msecs)
{
    long seconds = 0, microseconds = 0;
    struct timeval tmp, res;
    seconds = msecs / 1000;
    microseconds = (msecs % 1000) * 1000;
    tmp.tv_sec = seconds;
    tmp.tv_usec = microseconds;

    timeradd(t, &tmp, &res);
    memcpy(t, &res, sizeof(res));
    return;
}

/**
 * helper function for the sake of a cleaner interface :-)
 */
int adl_gettime(struct timeval *tv)
{
    return (gettimeofday(tv, (struct timezone *) NULL));
}

/**
 * function is to return difference in msecs between time a and b (i.e. a-b)
 * @param a later time (e.g. current time)
 * @param b earlier time
 * @return -1 if a is earlier than b, else msecs that passed from b to a
 */
int adl_timediff_to_msecs(struct timeval *a, struct timeval *b)
{
    struct timeval result;
    int retval;
    /* result = a-b */
    timersub(a, b, &result);
    retval = result.tv_sec * 1000 + result.tv_usec / 1000;
    event_logi(VVERBOSE, "Computed Time Difference : %d msecs", retval);
    return ((retval < 0) ? -1 : retval);
}

/**
 * function initializes the array of fds we want to use for listening to events
 * USE    POLL_FD_UNUSED to differentiate between used/unused fds !
 */
int init_poll_fds()
{
    int i;
    for (i = 0; i < NUM_FDS; i++) {
        assign_poll_fd(i, POLL_FD_UNUSED, 0);
    }
    num_of_fds = 0;
    return (0);
}


/**
 * 	function to check for events on all poll fds (i.e. open sockets), or else
 * 	execute the next timer event. Executed timer events are removed from the list.
 *  Wrapper to poll() -- returns after timeout or read event
 *	@return  0 after shutdown of the application, -1 for error
 *    @author  ajung
 */
int sctp_event_loop()
{
    int result;
    unsigned int u_res;
    int msecs;

    msecs = get_msecs_to_nexttimer();
    /* returns -1 if no timer in list */
    /* if (msecs > GRANULARITY || msecs < 0) */
    if (msecs < 0)
        msecs = GRANULARITY;
    if (msecs == 0) {
        dispatch_timer();
        return (0);
    }

    /*  print_debug_list(INTERNAL_EVENT_0); */
    result = poll(poll_fds, num_of_fds, msecs);
    switch (result) {
    case -1:
        error_log(ERROR_MAJOR, "Poll returned error, aborting !");
        return 0;
        break;
    case 0:
        dispatch_timer();
        break;
    default:
        u_res = (unsigned int) result;
        event_logi(INTERNAL_EVENT_0,
                   "############### %d Read Event(s) occurred -> dispatch_event()#############\n",
                   u_res);
        dispatch_event(result);
        break;
    }
    return (result);
}



void adl_init_adaptation_layer(gboolean support_ipv6, gint drop_frames)
{
    int uid = getuid();

    dropflg =  drop_frames;

    init_poll_fds();
    init_timer_list();
    /*  print_debug_list(INTERNAL_EVENT_0); */

    sctp_sfd = adl_open_sctp_socket(AF_INET);
    icmp_sfd = adl_open_icmp_socket();
#ifdef HAVE_IPV6
    if (support_ipv6 == TRUE) {
        sctpv6_sfd = adl_open_sctp_socket(AF_INET6);
        icmpv6_sfd = adl_open_icmpv6_socket();
        adl_register_fd_cb(icmpv6_sfd, POLLIN | POLLPRI | POLLERR, adl_icmpv6_cb);
    }
#endif
    /* now we can drop privileges */
    setuid(uid);
    adl_register_fd_cb(icmp_sfd, POLLIN | POLLPRI | POLLERR, adl_icmp_cb);
    event_log(INTERNAL_EVENT_0, "Registered callback adl_icmp_cb...\n");
    return;
}

/**
 * this function is supposed to open and bind a UDP socket listening on a port
 * to incoming udp pakets from localhost.
 * @param  support_ipv6 we may add support for udp via ipv6 here later ?!
 * @param  udp_port port this SCTP daemon is to listen to locally
 * @param  scf  callback funtion that is called when data has arrived
 * @return number of registered callbacks, or -1 if error ocurred
 */
int sctp_init_ulp_comm(gboolean support_ipv6, guint16 udp_port, Socket_callback_function scf)
{
    if (udpPort != 0) {
        error_log(ERROR_MAJOR, "UDP Callback has already been assigned - sorry, fix your program !");
        return -1;
    }
    ulp_sfd = -1;
    if (udp_port != 0) {
        /* save it in a static variable for further reference */
        udpPort = udp_port;
        ulp_sfd = adl_open_ulp_socket();
    }
    if (ulp_sfd != -1) {
        return (adl_register_fd_cb(ulp_sfd, POLLIN | POLLPRI, scf));
    }
    return -1;
}


/**
 * This function should only be called in exceptional cases...e.g. when we
 *  communicate with the ULP via an additional socket
 */
int
adl_register_socket_cb(gint sfd, Socket_callback_function scf)
{
    return (adl_register_fd_cb(sfd, POLLIN | POLLPRI, scf));
}


/**
 *      This function adds a callback that is to be called some time from now. It realizes
 *      the timer (in an ordered list).
 *      @param      milliseconds  action is to be started in milliseconds ms from now
 *      @param      action        pointer to a function to be executed, when timer goes off
 *      @return     returns an id value, that can be used to cancel a timer
 *      @author     ajung
 */
unsigned int
adl_start_timer(unsigned int milliseconds, Timer_expires_callback timer_cb,
                void *param1, void *param2)
{
    unsigned int result = 0;
    int res = 0;
    AlarmTimer item;
    struct timeval talarm;

#ifdef HAVE_GETTIMEOFDAY
    res = gettimeofday(&talarm, NULL);
#endif
    if (res < 0)
        error_log(ERROR_FATAL, "Gettimeofday returned an error in adl_start_timer !\n");
    adl_add_msecs_totime(&talarm, milliseconds);
    item.timer_type = 0;
    memcpy(&(item.action_time), &talarm, sizeof(struct timeval));
    item.action = timer_cb;
    item.arg1 = param1;
    item.arg2 = param2;
    result = insert_item(&item);
    return (result);
}

/**
 *      This function adds a callback that is to be called some time from now. It realizes
 *      the timer (in an ordered list).
 *      @param      tid        timer-id of timer to be removed
 *      @return     returns 0 on success, 1 if tid not in the list, -1 on error
 *      @author     ajung
 */
int adl_stop_timer(unsigned int tid)
{
    AlarmTimer tmp;
    return (remove_item(tid, &tmp));
}

/**
 *      Restarts a timer currently running
 *      @param      timer_id   the value returned by set_timer for a certain timer
 *      @param      milliseconds  action is to be started in milliseconds ms from now
 *      @return     new timer id , zero when there is an error (i.e. no timer)
 *      @author     ajung
 */
unsigned int adl_restart_timer_action(unsigned int timer_id, unsigned int milliseconds)
{
    unsigned int result;
    result = update_item(timer_id, milliseconds);
    event_logiii(VVERBOSE,
                 "Restarted Timer : timer_id = %u, msecs = %u, result = %u",
                 timer_id, milliseconds, result);
    return result;
}



/**
 *	function to close a bound socket from our list of socket descriptors
 *	@param	sfd	socket file descriptor to be closed
 *	@return  0 on success, -1 for error, 1 if socket was not bound
 *    @author  ajung
 */
int remove_cb(int sfd)
{
    int result;
    result = close(sfd);
    if (result < 0)
        error_log(ERROR_FATAL, "Close Socket resulted in an error");
    remove_poll_fd(sfd);
    return result;
}
