/*
 *  $Id: sctp.h,v 1.37 2001/01/04 18:13:15 ajung Exp $
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 2000 by Siemens AG, Munich, Germany.
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at www.sctp.de which should be used for
 * any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          Michael.Tuexen@icn.siemens.de
 *          ajung@exp-math.uni-essen.de
 *
 * Purpose: This header-file defines the interface for the ULP, the application
 *          that uses SCTP.
 *          Further comments on message distribution can be found in the implemen-
 *          tation of this interface distribution.c.
 * 
 * function prefix: sctp_         
 */

#ifndef SCTP_H
#define SCTP_H


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <sys/socket.h>
#include <netinet/in.h>
#include <glib.h>



/******************** Defines *********************************************************************/

/* Pathstatus, used with networkstatus primitives */
#define PATH_OK              0
#define PATH_UNREACHABLE     1

/******************** Structure Definitions *******************************************************/

/* union for handling either type of addresses: ipv4 and ipv6 */
#ifndef SOCKUNION_DEFINE
#define SOCKUNION_DEFINE    1
union sockunion
{
    struct sockaddr sa;
    struct sockaddr_in sin;
#ifdef HAVE_IPV6
    struct sockaddr_in6 sin6;
#endif                          /* HAVE_IPV6 */
};

#endif  /* SOCKUNION_DEFINE */


/* This struct containes the pointers to ULP callback functions.
   Each SCTP-instance can have its own set of callback functions.
   The callback functions of each SCTP-instance can be found by
   first reading the datastruct of an association from the list of 
   associations. The datastruct of the association contains the name
   of the SCTP instance to which it belongs. With the name of the SCTP-
   instance its datastruct can be read from the list of SCTP-instances.
*/

struct ULPcallback
{
    /*  indicates new data have arrived from peer (chapter 9.2.A).
       params: 1. associationID
       2. streamID
       3. length of data
     */
    void (*dataArriveNotif) (unsigned int, unsigned int, unsigned int, unsigned int);

    /* indicates a change of network status (chapter 9.2.C).
       params: 1.  associationID
       2.  destinationAddresses
       3.  newState 
     */
    void (*networkStatusChangeNotif) (unsigned int, short, unsigned short);

    /* indicates a sned failure (chapter 9.2.B).
       params: 1.  associationID
       2.  pointer to data not sent 
       3.  dataLength
       4.  context from sendChunk 
     */
    void (*sentFailureNotif) (unsigned int, unsigned char *, unsigned int, unsigned int *);

    /* indicates that communication was lost to peer (chapter 9.2.E).
       params: 1.  associationID
       2.  status, type of event
     */
    void (*communicationLostNotif) (unsigned int, unsigned short);

    /* indicates that a association is established (chapter 9.2.D).
       params: 1.  associationID
       2.  status, type of event
       3.  destination addresses
       4.  number of destination addresses
       5.  number input streamns
       6.  number output streams
     */
    void (*communicationUpNotif) (unsigned int, unsigned short,
                                  union sockunion *, int, unsigned short, unsigned short);
};


/*
 *  This struct contains the data to be returned to the ULP with the
 *  status primitive. Pointers point to arrays indexed by the number of
 *  destination addresses.
 */
typedef struct SCTP_Status
{
    gushort association_state;
    guint16 number_of_addresses;
    union sockunion primaryDestinationAddress;
    gushort primaryAddressIndex;
    union sockunion *destinationAddresses; /*  from SCTP-Control                   */
    /*  from PathManagement                 */
    gshort *path_states;        /*  PM_ACTIVE     0, PM_INACTIVE   1    */
    guint32 *SRTT;              /*  im msecs                            */
    guint32 *RTO;               /*  im msecs                            */

    guint32 currentReceiverWindowSize; /* from FlowControl                     */
    guint32 *currentCongestionWindowSize;
    guint32 noOfChunksInSendQueue;

    guint32 noOfChunksInRetransmissionQueue; /* from ReliableTransfer                */
    guint32 noOfChunksInReceptionQueue; /* from StreamEngine ?                  */
}
AssociationStatus;



/******************** Function Definitions ********************************************************/

unsigned short sctp_registerSCTP_instance(gboolean ulp_via_udp, gboolean with_ipv6, gint drop_packets,
                                          unsigned short localPort,
                                          unsigned short noOfInStreams,
                                          unsigned int noOfLocalAddresses,
                                          union sockunion *localAddressList,
                                          struct ULPcallback ULPcallbackFunctions);


gint32 sctp_unregisterSCTP_instance(unsigned short instance_name);


unsigned int sctp_associate(unsigned short SCTP_InstanceName,
                            unsigned short noOfOutStreams,
                            union sockunion *destinationAddress,
                            unsigned short destinationPort, gboolean useLocalPort);


int sctp_shutdown(unsigned int associationID);



int sctp_abort(unsigned int associationID);


int sctp_sendChunk(unsigned int associationID, unsigned short *streamID, /* optional (=null if none -> stream 0) */
                   unsigned char *buffer, unsigned int length, unsigned int protocolId, short *destAddressIndex, /* optional (=null if none) -> primary p. */
                   unsigned int *context, /* optional (=null if none) */
                   unsigned int *lifetime, /* optional (=null if none) -> infinit */
                   gboolean * unorderedDelivery, /* optional (=null if none) -> ordered */
                   gboolean * dontBundle); /* optional (=null if none) -> bundle */


short sctp_setPrimary(unsigned int associationID, short destAddressIndex);




unsigned short sctp_receive(unsigned int associationID,
                            unsigned short streamID, unsigned char *buffer, unsigned int *length);


AssociationStatus *sctp_status(unsigned int associationID);


int sctp_changeHeartBeat(unsigned int associationID,
                         short destAddressIndex, gboolean heartbeatON, unsigned int timeIntervall);


int sctp_requestHeartbeat(unsigned int associationID, short destAddressIndex);


unsigned int sctp_getSRTT_Report(unsigned int associationID, short destAddressIndex);

int sctp_setFailureThreshold(unsigned int associationID,
                             short destAddressIndex, unsigned short pathMaxRetransmissions);


int sctp_setProtocolParams(unsigned int associationID, short destAddressIndex, /* optional (=null if none) */
                           unsigned int RTO_initial,
                           unsigned int validCookieLife, unsigned int assocMaxRetransmits, unsigned int pathMaxRetransmits, /* per path */
                           unsigned int maxInitRetransmits);


unsigned int sctp_get_assocID(union sockunion * destination);



/* Defines the callback function that is called when new data have been received
   (Events POLLIN||POLLPRI )
   Params: 1. socketdescriptor of the socket where the datagramm has been received
           2. pointer to the datagramm data
           3. length of datagramm data
           4. source Address
           5. length of source address
*/
typedef void (*Socket_callback_function) (gint, unsigned char *,
                                          int, union sockunion *, int);

/**
 * this function is supposed to open and bind a UDP socket listening on a port
 * to incoming udp pakets from localhost.
 * @param  support_ipv6 we may add support for udp via ipv6 here later ?!
 * @param  udp_port port this SCTP daemon is to listen to locally
 * @param  scf  callback funtion that is called when data has arrived
 * @return number of registered callbacks, or -1 if error ocurred
 */
int sctp_init_ulp_comm(gboolean support_ipv6, guint16 udp_port, Socket_callback_function scf);

/* Defines the callback function that is called when an timer expires.
   Params: 1. ID of timer
           2. pointer to param1
           3. pointer to param2
              param1 and param2 are transparent data from the caller of start_timer,
              that are returned to it when the timer expires.
*/
typedef void (*Timer_expires_callback) (unsigned int, void *, void *);


/**
 *      This function adds a callback that is to be called some time from now. It realizes
 *      the timer (in an ordered list).
 *      @param      milliseconds  action is to be started in milliseconds ms from now
 *      @param      timer_cb      pointer to a function to be executed, when timer expires
 *      @param      param1        pointer to be returned to the caller when timer expires
 *      @param      param2        pointer to be returned to the caller when timer expires
 *      @return     returns an id value, that can be used to cancel a timer
 *      @author     ajung
 */
unsigned int adl_start_timer(unsigned int milliseconds,
                        Timer_expires_callback timer_cb, void *param1, void *param2);



/**
 *      This function stops a previously started timer.
 *      @param      tid        timer-id of timer to be removed
 *      @return     returns 0 on success, 1 if tid not in the list, -1 on error
 *      @author     ajung
 */
int adl_stop_timer(unsigned int tid);

/**
 *      Restarts a timer currently running
 *      @param      timer_id   the value returned by set_timer for a certain timer
 *      @param      milliseconds  action is to be started in milliseconds ms from now
 *      @return     new timer id , zero when there is an error (i.e. no timer)
 *      @author     ajung
 */
unsigned int adl_restart_timer_action(unsigned int timer_id, unsigned int milliseconds);

/**
 * helper function for the sake of a cleaner interface :-)
 */
int adl_gettime(struct timeval *tv);


int sctp_event_loop();


#endif
